# mav_controller.py
# version: 1.0.1
# Author: Theodore Tasman
# Date: 2025-01-30
# Organization: PSU UAS

"""
This module is responsible for controlling ardupilot.
"""

from pymavlink import mavutil
from MAVez.coordinate import Coordinate


class Controller:
    """
    Controller class for atomic MAVLink communication with ardupilot.

    Args:
        connection_string (str): The connection string for ardupilot. Default is "tcp:127.0.0.1:5762" used for SITL.
        baud (int): The baud rate for the connection. Default is 57600.
        logger: Logger instance for logging messages (optional).

    Raises:
        ConnectionError: If the connection to ardupilot fails.

    Returns:
        A Controller instance that can be used to communicate with ardupilot.
    """

    # class wide error codes
    # if an error is function specific, it will be defined in the function and documented in class docstring
    TIMEOUT_ERROR = 101

    TIMEOUT_DURATION = 5  # timeout duration in seconds

    def __init__(self, connection_string="tcp:127.0.0.1:5762", baud=57600, logger=None):
        """
        Initialize the controller.

        Args:
            connection_string (str): The connection string for ardupilot.
            baud (int): The baud rate for the connection.
            logger: Logger instance for logging messages (optional).

        Raises:
            ConnectionError: If the connection to ardupilot fails.

        Returns:
            None
        """
        self.logger = logger

        self.master = mavutil.mavlink_connection(connection_string, baud=baud)

        response = self.master.wait_heartbeat(
            blocking=True, timeout=self.TIMEOUT_DURATION
        )

        # check if the connection was successful
        if not response:
            if self.logger:
                self.logger.error("[Controller] Connection failed")
            raise ConnectionError("Connection failed")

    def set_logger(self, logger):
        """
        Set the logger for the controller.

        Args:
            logger: Logger instance for logging messages.

        Returns:
            None
        """
        self.logger = logger

    def decode_error(self, error_code):
        """
        Decode the error code into a human-readable string.

        Args:
            error_code (int): The error code to decode.

        Returns:
            str: A human-readable error message.
        """
        errors_dict = {
            101: "\nRESPONSE TIMEOUT ERROR (101)\n",
            111: "\nUNKNOWN MODE ERROR (102)\n",
        }

        return errors_dict.get(error_code, f"UNKNOWN ERROR ({error_code})")

    def await_mission_request(self):
        """
        Wait for a mission request from ardupilot.

        Returns:
            int: Mission index if a mission request was received, 101 if the response timed out.
        """

        # MISSION_REQUEST IS DEPRECATED, but likely still used. (replaced by MISSION_REQUEST_INT)
        # Always respond with MISSION_ITEM_INT.
        # If this stops working, try receiving MISSION_REQUEST_INT instead.
        response = self.master.recv_match(
            type="MISSION_REQUEST", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received mission request: {response.seq}"
                )
            return response.seq
        else:
            if self.logger:
                self.logger.error("[Controller] Mission request timed out")
            return self.TIMEOUT_ERROR

    def await_mission_ack(self):
        """
        Wait for a mission ack from ardupilot.

        Returns:
            int: 0 if the mission ack was received, error code if there was an error, 101 if the response timed out.
        """

        response = self.master.recv_match(
            type="MISSION_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.type == 0:
                if self.logger:
                    self.logger.info("[Controller] Received mission ack.")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Mission ack error: {response.type}"
                    )
                return response.type
        else:
            if self.logger:
                self.logger.error("[Controller] Mission ack timed out")
            return self.TIMEOUT_ERROR

    def send_message(self, message):
        """
        Send a MAVLink message to ardupilot.

        Args:
            message: The MAVLink message to send.

        Returns:
            None
        """
        self.master.mav.send(message)

    def send_mission_count(self, count, mission_type=0):
        """
        Send the mission count to ardupilot.

        Args:
            count (int): The number of mission items.
            mission_type (int): The type of mission (default is 0 for MISSION_TYPE 0).

        Returns:
            int: 0 if the mission count was sent successfully, 101 if the response timed out.
        """

        self.master.mav.mission_count_send(
            0,  # target_system
            0,  # target_component
            count,  # count
            mission_type,  # mission_type
        )
        if self.logger:
            self.logger.info(f"[Controller] Sent mission count: {count}")
        return 0

    def await_mission_item_reached(self, timeout=TIMEOUT_DURATION):
        """
        Wait for a mission item reached message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            int: The sequence number of the reached mission item if received, TIMEOUT_ERROR (101) if the response timed out.
        """

        response = self.master.recv_match(
            type="MISSION_ITEM_REACHED", blocking=True, timeout=timeout
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received mission item reached: {response.seq}"
                )
            return response.seq
        else:
            if self.logger:
                self.logger.error("[Controller] Mission item reached timed out")
            return self.TIMEOUT_ERROR

    def send_clear_mission(self):
        """
        Clear the mission on ardupilot.

        Returns:
            int: 0 if the mission was cleared successfully
        """

        self.master.waypoint_clear_all_send()
        if self.logger:
            self.logger.info("[Controller] Sent clear mission")
        return 0

    def set_mode(self, mode):
        """
        Set the ardupilot mode.

        Args:
            mode (str): The mode to set ardupilot to. Options include: "AUTO", "GUIDED", "FBWA", etc...

        Returns:
            int: 0 if the mode was set successfully, 111 if the mode is unknown, 101 if the response timed out.
        """
        UNKNOWN_MODE = 111

        if mode not in self.master.mode_mapping():
            if self.logger:
                self.logger.error(f"[Controller] Unknown mode: {mode}")
            return UNKNOWN_MODE

        mode_id = self.master.mode_mapping()[mode]
        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_DO_SET_MODE,  # command
            0,  # confirmation
            mavutil.mavlink.MAV_MODE_FLAG_CUSTOM_MODE_ENABLED,  # param1
            mode_id,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(f"[Controller] Set mode to {mode}")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to set mode: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Set mode timed out")
            return self.TIMEOUT_ERROR

    def arm(self, force=False):
        """
        Arm ardupilot

        Args:
            force (bool): If True, ardupilot will be armed regardless of its state.

        Returns:
            int: 0 if ardupilot was armed successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_COMPONENT_ARM_DISARM,  # command
            0,  # confirmation
            1,  # param1
            21196 if force else 0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info("[Controller] Drone armed successfully")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to arm drone: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Arm command timed out")
            return self.TIMEOUT_ERROR

    def disarm(self, force=False):
        """
        Disarm ardupilot.

        Args:
            force (bool): If True, ardupilot will be disarmed regardless of its state.

        Returns:
            int: 0 if ardupilot was disarmed successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_COMPONENT_ARM_DISARM,  # command
            0,  # confirmation
            0,  # param1
            21196 if force else 0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info("[Controller] Drone disarmed successfully")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to disarm drone: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Disarm command timed out")
            return self.TIMEOUT_ERROR

    def enable_geofence(self):
        """
        Enable the geofence.

        Returns:
            int: 0 if the geofence was enabled successfully, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_DO_FENCE_ENABLE,  # command
            0,  # confirmation
            1,  # param1
            0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info("[Controller] Geofence enabled successfully")
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to enable geofence: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Geofence enable command timed out")
            return self.TIMEOUT_ERROR

    def disable_geofence(self, floor_only=False):
        """
        Disable the geofence.

        Args:
            floor_only (bool): If True, only the floor of the geofence will be disabled.

        Returns:
            int: 0 if the geofence was disabled successfully, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_DO_FENCE_ENABLE,  # command
            0,  # confirmation
            2 if floor_only else 0,  # param1
            0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info("[Controller] Geofence disabled successfully")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to disable geofence: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Geofence disable command timed out")
            return self.TIMEOUT_ERROR

    def set_home(self, home_coordinate=Coordinate(0, 0, 0)):
        """
        Set the home location.

        Args:
            home_coordinate (Coordinate): The home coordinate to set. If the coordinate is (0, 0, 0), the current GPS location will be used.

        Returns:
            int: 0 if the home location was set successfully, error code if there was an error, 101 if the response timed out.
        """

        # use_current is set to True if the home coordinate is (0, 0, 0)
        use_current = home_coordinate == (0, 0, 0)
        # if alt is 0, use the current altitude
        home_coordinate.alt = (
            self.receive_gps().alt if home_coordinate.alt == 0 else home_coordinate.alt
        )

        message = self.master.mav.command_int_encode(
            0,  # target_system
            0,  # target_component
            0,  # frame - MAV_FRAME_GLOBAL
            mavutil.mavlink.MAV_CMD_DO_SET_HOME,  # command
            0,  # current
            0,  # auto continue
            1 if use_current else 0,  # param1
            0,  # param2
            0,  # param3
            0,  # param4
            home_coordinate.lat,  # param5
            home_coordinate.lon,  # param6
            int(home_coordinate.alt),  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(
                        f"[Controller] Home location set to {home_coordinate}"
                    )
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to set home location: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Set home location command timed out")
            return self.TIMEOUT_ERROR

    def set_servo(self, servo_number, pwm):
        """
        Set the a servo to a specified PWM value.

        Args:
            servo_number (int): The servo number to set.
            pwm (int): The PWM value to set the servo to.

        Returns:
            int: 0 if the servo was set successfully, error code if there was an error, 101 if the response timed out.
        """
        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_DO_SET_SERVO,  # command
            0,  # confirmation
            servo_number,  # param1
            pwm,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)
        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(f"[Controller] Set servo {servo_number} to {pwm}")
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to set servo {servo_number}: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Set servo command timed out")
            return self.TIMEOUT_ERROR

    def receive_channel_input(self, timeout=TIMEOUT_DURATION):
        """
        Wait for an RC_CHANNELS message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            response if an RC_CHANNELS message was received, 101 if the response timed out
        """

        response = self.master.recv_match(
            type="RC_CHANNELS", blocking=True, timeout=timeout
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received channel input from {response.chancount} channels"
                )
            return response
        else:
            if self.logger:
                self.logger.error("[Controller] Receive channel input timed out")
            return self.TIMEOUT_ERROR

    def receive_wind(self, timeout=TIMEOUT_DURATION):
        """
        Wait for a wind_cov message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            response if a wind_cov message was received, 101 if the response timed out
        """
        response = self.master.recv_match(
            type="WIND_COV", blocking=True, timeout=timeout
        )
        if response:
            if self.logger:
                self.logger.info("[Controller] Received wind data")
            return response
        else:
            if self.logger:
                self.logger.error("[Controller] Receive wind data timed out")
            return self.TIMEOUT_ERROR

    def receive_gps(self, timeout=TIMEOUT_DURATION):
        """
        Wait for a gps_raw_int message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            Coordinate: A Coordinate object containing the GPS data if received, TIMEOUT_ERROR (101) if the response timed out.
        """
        response = self.master.recv_match(
            type="GLOBAL_POSITION_INT", blocking=True, timeout=timeout
        )

        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received GPS data, lat: {response.lat}, lon: {response.lon}, alt: {response.alt / 1000}, heading: {response.hdg}"
                )
            return Coordinate(
                response.lat,
                response.lon,
                response.alt / 1000,
                use_int=False,
                heading=response.hdg,
            )  # convert to meters, lat and lon are in degrees e7

        else:
            if self.logger:
                self.logger.error("[Controller] Receive GPS data timed out")
            return self.TIMEOUT_ERROR

    def receive_landing_status(self, timeout=TIMEOUT_DURATION):
        """
        Wait for a landed_state message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            int: The landing state if received, TIMEOUT_ERROR (101) if the response timed out, 0 if the state is undefined, 1 if on ground, 2 if in air, 3 if taking off, 4 if landing.
        """
        response = self.master.recv_match(
            type="EXTENDED_SYS_STATE", blocking=True, timeout=timeout
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received landing status: {response.landed_state}"
                )
            return response.landed_state
        else:
            if self.logger:
                self.logger.error("[Controller] Receive landing status timed out")
            return self.TIMEOUT_ERROR

    def set_message_interval(self, message_type, interval, timeout=TIMEOUT_DURATION):
        """
        Set the message interval for the specified message type.

        Args:
            message_type (str): The type of message to set the interval for.
            interval (int): The interval in milliseconds to set for the message type.
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            int: 0 if the message interval was set successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_SET_MESSAGE_INTERVAL,  # command
            0,  # confirmation
            message_type,  # param1
            interval,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=timeout
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(
                        f"[Controller] Set message interval for {message_type} to {interval}"
                    )
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to set message interval: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Set message interval command timed out")
            return self.TIMEOUT_ERROR

    def disable_message_interval(self, message_type, timeout=TIMEOUT_DURATION):
        """
        Disable the message interval for the specified message type.

        Args:
            message_type (str): The type of message to disable the interval for.
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            int: 0 if the message interval was disabled successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_SET_MESSAGE_INTERVAL,  # command
            0,  # confirmation
            message_type,  # param1
            -1,  # param2 # -1 disables the message
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=timeout
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(
                        f"[Controller] Disabled message interval for {message_type}"
                    )
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to disable message interval: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error(
                    "[Controller] Disable message interval command timed out"
                )
            return self.TIMEOUT_ERROR

    def await_current_mission_index(self):
        """
        Get the current mission index.

        Returns:
            int: The current mission index if received, TIMEOUT_ERROR (101) if the response timed out.
        """

        response = self.master.recv_match(
            type="MISSION_CURRENT", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received current mission index: {response.seq}"
                )
            return response.seq
        else:
            if self.logger:
                self.logger.error(
                    "[Controller] Receive current mission index timed out"
                )
            return self.TIMEOUT_ERROR

    def set_current_mission_index(self, index):
        """
        sets the target mission index to the specified index

        Args:
            index (int): The index to set as the current mission index.

        Returns:
            int: 0 if the current mission index was set successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_DO_SET_MISSION_CURRENT,  # command
            0,  # confirmation
            index,  # param1
            0,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )
        self.master.mav.send(message)
        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(
                        f"[Controller] Set current mission index to {index}"
                    )
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to set current mission index: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error(
                    "[Controller] Set current mission index command timed out"
                )
            return self.TIMEOUT_ERROR

    def start_mission(self, start_index, end_index):
        """
        Start the mission at the specified index.

        Args:
            start_index (int): The index to start the mission from.
            end_index (int): The index to end the mission at.

        Returns:
            int: 0 if the mission was started successfully, error code if there was an error, 101 if the response timed out.
        """

        message = self.master.mav.command_long_encode(
            0,  # target_system
            0,  # target_component
            mavutil.mavlink.MAV_CMD_MISSION_START,  # command
            0,  # confirmation
            start_index,  # param1
            end_index,  # param2
            0,  # param3
            0,  # param4
            0,  # param5
            0,  # param6
            0,  # param7
        )

        self.master.mav.send(message)

        response = self.master.recv_match(
            type="COMMAND_ACK", blocking=True, timeout=self.TIMEOUT_DURATION
        )
        if response:
            if response.result == 0:
                if self.logger:
                    self.logger.info(
                        f"[Controller] Started mission from {start_index} to {end_index}"
                    )
                return 0
            else:
                if self.logger:
                    self.logger.error(
                        f"[Controller] Failed to start mission: {response.result}"
                    )
                return response.result
        else:
            if self.logger:
                self.logger.error("[Controller] Start mission command timed out")
            return self.TIMEOUT_ERROR

    def receive_attitude(self, timeout=TIMEOUT_DURATION):
        """
        Wait for an attitude message from ardupilot.

        Args:
            timeout (int): The timeout duration in seconds. Default is 5 seconds.

        Returns:
            response if an attitude message was received, TIMEOUT_ERROR (101) if the response timed out.
        """

        response = self.master.recv_match(
            type="ATTITUDE", blocking=True, timeout=timeout
        )
        if response:
            if self.logger:
                self.logger.info(
                    f"[Controller] Received attitude data: roll={response.roll}, pitch={response.pitch}, yaw={response.yaw}"
                )
            return response
        else:
            if self.logger:
                self.logger.error("[Controller] Receive attitude data timed out")
            return self.TIMEOUT_ERROR



# flight_controller.py
# version: 1.0.1
# Original Author: Theodore Tasman
# Date: 2025-01-30
# Organization: PSU UAS

"""
This module is responsible for managing the flight of ardupilot.
"""

# SITL Start Command:
# python3 ./MAVLink/ardupilot/Tools/autotest/sim_vehicle.py -v ArduPlane --console --map --custom-location 38.31527628,-76.54908330,40,282.5

from MAVez.coordinate import Coordinate
from MAVez.mission import Mission
from MAVez.controller import Controller
import time


class Flight_Controller(Controller):
    """
    Manages the flight plan for ardupilot. Extends the Controller class to provide complex flight functionalities.

    Args:
        connection_string (str): The connection string to ardupilot.
        logger (Logger | None): Optional logger for logging flight events.

    Raises:
        ConnectionError: If the connection to ardupilot fails.

    Returns:
        Flight_Controller: An instance of the Flight_Controller class.
    """

    PREFLIGHT_CHECK_ERROR = 301
    DETECT_LOAD_ERROR = 302
    AIRDROP_NOT_BUILT_ERROR = 303

    def __init__(self, connection_string="tcp:127.0.0.1:5762", logger=None):
        # Initialize the controller
        super().__init__(connection_string, logger=logger)

        # initialize preflight check
        self.preflight_check_done = False

        # create missions
        self.takeoff_mission = Mission(self)
        self.detect_mission = Mission(self)
        self.land_mission = Mission(self)
        self.airdrop_mission = Mission(self)
        self.geofence = Mission(self, type=1)  # type 1 is geofence

        # initialize mission list
        self.mission_list = [self.takeoff_mission]  # TODO: takeoff mission

    def decode_error(self, error_code):
        """
        Decode an error code.

        Args:
            error_code (int): The error code to decode.

        Returns:
            str: A string describing the error.
        """

        errors_dict = {
            101: "\nTIMEOUT ERROR (101)\n",
            111: "\nUNKNOWN MODE ERROR (102)\n",
            301: "\nPREFLIGHT CHECK ERROR (301)\n",
            302: "\nDETECT LOAD ERROR (302)\n",
        }

        return errors_dict.get(error_code, f"UNKNOWN ERROR ({error_code})")

    def takeoff(self, takeoff_mission_file):
        """
        Takeoff ardupilot. Preflight check must be done first.

        Args:
            takeoff_mission_file (str): The file containing the takeoff mission.

        Returns:
            int: 0 if the takeoff was successful, otherwise an error code.
        """

        # verify preflight check
        if not self.preflight_check_done:
            return self.PREFLIGHT_CHECK_ERROR

        # Load the takeoff mission from the file
        response = self.takeoff_mission.load_mission_from_file(takeoff_mission_file)

        # verify that the mission was loaded successfully
        if response:
            if self.logger:
                self.logger.critical("[Flight] Takeoff failed, mission not loaded")
            return response

        # send the takeoff mission
        response = self.takeoff_mission.send_mission()

        # verify that the mission was sent successfully
        if response:
            if self.logger:
                self.logger.critical("[Flight] Takeoff failed, mission not sent")
            return response

        # wait for mission to be fully received
        # Countdown from 5
        if self.logger:
            self.logger.info("[Flight] Takeoff in 5 seconds")
        for i in range(5, 0, -1):
            time.sleep(1)

        # set the mode to AUTO
        response = self.set_mode("AUTO")

        # verify that the mode was set successfully
        if response:
            if self.logger:
                self.logger.critical("[Flight] Takeoff failed, mode not set to AUTO")
            return response

        # arm ardupilot
        response = self.arm()

        # verify that ardupilot was armed successfully
        if response:
            if self.logger:
                self.logger.critical("[Flight] Takeoff failed, drone not armed")
            return response

        return 0

    def append_mission(self, filename):
        """
        Append a mission to the mission list.

        Args:
            filename (str): The file containing the mission to append.

        Returns:
            int: 0 if the mission was appended successfully, otherwise an error code.
        """
        # Load the mission from the file
        mission = Mission(self)
        result = mission.load_mission_from_file(filename)

        if result:
            if self.logger:
                self.logger.critical("[Flight] Could not append mission.")
            return result

        if self.logger:
            self.logger.info(
                f"[Flight] Appended mission from {filename} to mission list"
            )
        self.mission_list.append(mission)
        return 0

    def wait_for_waypoint_reached(self, target, timeout=30):
        """
        Wait for ardupilot to reach the current waypoint.

        Args:
            target (int): The target waypoint index to wait for.
            timeout (int): The maximum time to wait for the waypoint to be reached in seconds.

        Returns:
            int: 0 if the waypoint was reached successfully, otherwise an error code.
        """
        latest_waypoint = -1

        if self.logger:
            self.logger.info(f"[Flight] Waiting for waypoint {target} to be reached")

        while latest_waypoint < target:
            response = self.await_mission_item_reached(timeout)

            if response == self.TIMEOUT_ERROR:
                return response

            latest_waypoint = response

        if self.logger:
            self.logger.info(f"[Flight] Waypoint {target} reached")
        return 0

    def wait_and_send_next_mission(self):
        """
        Waits for the last waypoint to be reached, clears the mission, sends the next mission, sets mode to auto.

        Returns:
            int: 0 if the next mission was sent successfully, otherwise an error code.
        """
        # Get the current mission
        current_mission = self.mission_list.pop(0)

        # if the mission list is empty, return
        if len(self.mission_list) == 0:
            if self.logger:
                self.logger.info("[Flight] No more missions in list, landing")
            # next_mission = self.land_mission
            return 0

        # otherwise, set the next mission to the next mission in the list
        else:
            if self.logger:
                self.logger.info(
                    f"[Flight] Queuing next mission in list of {len(self.mission_list)} missions"
                )
            next_mission = self.mission_list[0]

        # calculate the target index
        target_index = len(current_mission) - 1

        # Wait for the target index to be reached
        response = self.wait_for_waypoint_reached(target_index, 60)

        # verify that the response was received
        if response == self.TIMEOUT_ERROR:
            if self.logger:
                self.logger.critical("[Flight] Failed to wait for next mission.")
            return response

        # Clear the mission
        response = current_mission.clear_mission()
        if response:
            if self.logger:
                self.logger.critical("[Flight] Failed to send next mission.")
            return response

        # Send the next mission
        result = next_mission.send_mission()
        if result:
            if self.logger:
                self.logger.critical("[Flight] Failed to send next mission.")
            return result

        # set the mode to AUTO
        response = self.set_mode("AUTO")

        # verify that the mode was set successfully
        if response:
            if self.logger:
                self.logger.critical("[Flight] Failed to send next mission.")
            return response

        if self.logger:
            self.logger.info("[Flight] Next mission sent")
        return result

    def wait_for_landed(self, timeout=60):
        """
        Wait for ardupilot to signal landed.

        Args:
            timeout (int): The maximum time to wait for the landing status in seconds.

        Returns:
            int: 0 if the landing was successful, otherwise an error code.
        """
        landing_status = -1

        # start receiving landing status
        response = self.set_message_interval(
            message_type=245, interval=1e6
        )  # 245 is landing status (EXTENDED_SYS_STATE), 1e6 is 1 second
        if response:
            if self.logger:
                self.logger.critical("[Flight] Failed waiting for landing.")
            return response

        # wait for landing status to be landed
        start_time = time.time()
        while (
            landing_status != 1
        ):  # 1 for landed, 2 for in air, 3 for taking off, 4 for currently landing, 0 for unknown
            # check for timeout
            if time.time() - start_time > timeout:
                response = self.TIMEOUT_ERROR
                if self.logger:
                    self.logger.error("[Flight] Timed out waiting for landing.")
                return response

            # get the landing status
            response = self.receive_landing_status()

            # verify that the response was received
            if response == self.TIMEOUT_ERROR:
                if self.logger:
                    self.logger.error("[Flight] Failed waiting for landing.")
                return response

            landing_status = response

        # stop receiving landing status
        response = self.disable_message_interval(
            message_type=245
        )  # 245 is landing status (EXTENDED_SYS_STATE)
        if response:
            if self.logger:
                self.logger.error("[Flight] Error waiting for landing.")
            return response

        return 0

    def preflight_check(
        self, land_mission_file, geofence_file, home_coordinate=Coordinate(0, 0, 0)
    ):
        """
        Perform a preflight check. On success, set preflight_check_done to True. Loads the land mission and geofence mission, sets home location, and enables geofence.

        Args:
            land_mission_file (str): Path to the file containing the land mission.
            geofence_file (str): Path to the file containing the geofence mission.
            home_coordinate (Coordinate): The home coordinate to set (optional).

        Returns:
            int: 0 if the preflight check passed, otherwise an error code.
        """

        # Set home location
        response = self.set_home(home_coordinate)

        # verify that the home location was set successfully
        if response:
            if self.logger:
                self.logger.critical(
                    "[Flight] Preflight check failed, home location not set"
                )
            return response

        # load geofence
        response = self.geofence.load_mission_from_file(geofence_file)

        # verify that the geofence was loaded successfully
        if response:
            if self.logger:
                self.logger.critical(
                    "[Flight] Preflight check failed, geofence not loaded"
                )
            return response

        # send geofence
        response = self.geofence.send_mission()

        # verify that the geofence was sent successfully
        if response:
            if self.logger:
                self.logger.critical(
                    "[Flight] Preflight check failed, geofence not sent"
                )
            return response

        # load land mission
        response = self.land_mission.load_mission_from_file(land_mission_file)

        # verify that the land mission was loaded successfully
        if response:
            if self.logger:
                self.logger.critical(
                    "[Flight] Preflight check failed, land mission not loaded"
                )
            return response

        # enable geofence
        response = self.enable_geofence()

        # verify that the geofence was enabled successfully
        if response:
            if self.logger:
                self.logger.critical(
                    "[Flight] Preflight check failed, geofence not enabled"
                )
            return response

        # set preflight check done
        self.preflight_check_done = True

        if self.logger:
            self.logger.info("[Flight] Preflight check passed")
        return 0

    def jump_to_next_mission_item(self):
        """
        Jump to the next mission item.

        Returns:
            int: 0 if the jump was successful, otherwise an error code.
        """

        if self.logger:
            self.logger.info("[Flight] Waiting for current mission index")
        # wait for the current mission target to be received (should be broadcast by default)
        response = self.await_current_mission_index()
        if response == self.TIMEOUT_ERROR:
            return response

        # jump to the next mission item
        response = self.set_current_mission_index(response + 1)
        if response:
            return response

        return 0

    def wait_for_channel_input(
        self, channel, value, timeout=10, wait_time=120, value_tolerance=100
    ):
        """
        Wait for a specified rc channel to reach a given value

        Args:
            channel (int): The channel number to wait for.
            value (int): The value to wait for.
            timeout (int): The maximum time to wait for an update on the channel in seconds.
            wait_time (int): The maximum time to wait for the channel to be set in seconds.
            value_tolerance (int): The tolerance range for the set value.

        Returns:
            int: 0 if the channel was set to the desired value, otherwise an error code
        """
        latest_value = -float("inf")
        start_time = time.time()

        # set the channel to be received
        channel = f"chan{channel}_raw"

        if self.logger:
            self.logger.info(
                f"[Flight] Waiting for channel {channel} to be set to {value}"
            )
        # only wait for the channel to be set for a certain amount of time
        while time.time() - start_time < wait_time:
            # get channel inputs
            response = self.receive_channel_input(timeout)

            # verify that the response was received
            if response == self.TIMEOUT_ERROR:
                if self.logger:
                    self.logger.critical("[Flight] Failed waiting for channel input.")
                return response

            # channel key is 'chanX_raw' where X is the channel number
            latest_value = getattr(response, channel)

            # print(f'Latest value: {latest_value}')
            # check if the value is within the tolerance range
            if (
                latest_value > value - value_tolerance
                and latest_value < value + value_tolerance
            ):
                if self.logger:
                    self.logger.info(
                        f"[Flight] Channel {channel} set to {latest_value}"
                    )
                return 0

        if self.logger:
            self.logger.critical(
                f"[Flight] Timed out waiting for channel {channel} to be set to {value}"
            )
        return self.TIMEOUT_ERROR

    def get_altitude(self):
        """
        Get the altitude from ardupilot.

        Returns:
            float: The altitude in meters if successful, otherwise an error code.
        """
        # get the altitude
        if self.logger:
            self.logger.info("[Flight] Getting altitude")

        response = self.receive_altitude()

        # verify that the response was received
        if response == self.TIMEOUT_ERROR:
            if self.logger:
                self.logger.critical("[Flight] Failed to get altitude.")
            return response

        return response



# coordinate.py
# version: 1.0.0
# Author: Theodore Tasman
# Date: 2025-01-30
# Organization: PSU UAS

"""
Represents a geographic coordinate with latitude, longitude, and altitude.
"""

import math

EARTH_RADIUS = 6378137  # in meters
METERS_PER_DEGREE = EARTH_RADIUS / 180 * math.pi


class Coordinate:
    """
    This class represents a coordinate in latitude, longitude, and altitude. It provides methods to initialize the coordinate, convert between degrees and decimal degrees, and offset the coordinate by a given distance and heading.

    Args:
        lat (float | str): Latitude in decimal degrees or DMS format if `dms` is True.
        lon (float | str): Longitude in decimal degrees or DMS format if `dms` is True.
        alt (float): Altitude in meters.
        dms (bool): If True, the coordinates are in degrees, minutes, seconds format. Defaults to False.
        use_int (bool): If True, the coordinates are stored as integers. Defaults to True.
        heading (float | None): Heading in degrees. Defaults to None.

    Returns:
        Coordinate: An instance of the Coordinate class.

    """

    def __init__(self, lat, lon, alt, dms=False, use_int=True, heading=None):
        # if dms is True, convert the coordinates to decimal degrees
        if dms:
            self.lat = self.dms_to_dd(lat)
            self.lon = self.dms_to_dd(lon)
        # otherwise, use the coordinates as they are
        else:
            self.lat = lat
            self.lon = lon

        # alt is always used as it is
        self.alt = alt

        # if integer is True, convert the coordinates to integers
        if use_int:
            self.lat = int(self.lat * 1e7)
            self.lon = int(self.lon * 1e7)

        self.is_int = use_int

        self.heading = heading

    def __str__(self):
        return (
            f"{self.lat},{self.lon},{self.alt},{self.heading}"
            if self.heading is not None
            else f"{self.lat},{self.lon},{self.alt}"
        )

    __repr__ = __str__

    def dms_to_dd(self, dms):
        """
        Convert degrees, minutes, seconds to decimal degrees.

        Args:
            dms (tuple): A tuple of (degrees, minutes, seconds).

        Returns:
            float: The decimal degrees equivalent of the DMS input.
        """
        dd = dms[0] + dms[1] / 60 + dms[2] / 3600
        return dd

    def offset_coordinate(self, offset, heading):
        """
        Offset the coordinate by a given distance and heading.

        Args:
            offset (float): The distance to offset in meters.
            heading (float): The heading in degrees.

        Returns:
            Coordinate: A new Coordinate object with the offset applied.
        """

        lat, lon = self.normalize()

        # convert heading to radians
        heading = math.radians(heading)

        # calculate the offset in latitude and longitude
        lat_offset = offset * math.cos(heading)
        lon_offset = offset * math.sin(heading)

        # convert the offset to degrees
        lat_offset = lat_offset / METERS_PER_DEGREE
        lon_offset = lon_offset / (METERS_PER_DEGREE * math.cos(math.radians(lat)))

        # calculate the new latitude and longitude
        new_lat = lat + lat_offset
        new_lon = lon + lon_offset

        return Coordinate(new_lat, new_lon, self.alt, use_int=self.is_int)

    def __eq__(self, other):
        if isinstance(other, tuple) and len(other) == 3:
            # if other is a tuple, convert it to a Coordinate
            other = Coordinate(other[0], other[1], other[2])

        return self.lat == other.lat and self.lon == other.lon and self.alt == other.alt

    def normalize(self):
        """
        Normalize the coordinates to decimal degrees.

        Returns:
            tuple: A tuple containing the latitude and longitude in decimal degrees.
        """
        # ensure both self and other are in decimal degrees
        if self.is_int:
            self_lat = self.lat / 1e7
            self_lon = self.lon / 1e7
        else:
            self_lat = self.lat
            self_lon = self.lon

        return self_lat, self_lon

    def distance_to(self, other):
        """
        Calculate the distance between two coordinates in meters using the haversine formula.

        Args:
            other (Coordinate): The other coordinate to calculate the distance to.

        Returns:
            float: The distance in meters between the two coordinates.
        """
        # ensure other is a Coordinate
        if not isinstance(other, Coordinate):
            raise TypeError("other must be a Coordinate")

        self_lat, self_lon = self.normalize()
        other_lat, other_lon = other.normalize()

        dlat = math.radians(other_lat - self_lat)
        dlon = math.radians(other_lon - self_lon)

        a = (
            math.sin(dlat / 2) ** 2
            + math.cos(math.radians(self_lat))
            * math.cos(math.radians(other_lat))
            * math.sin(dlon / 2) ** 2
        )
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        distance = EARTH_RADIUS * c
        return distance

    def bearing_to(self, other):
        """
        Calculate the bearing between two coordinates in degrees.

        Args:
            other (Coordinate): The other coordinate to calculate the bearing to.

        Returns:
            float: The bearing in degrees from self to other.
        """
        # ensure other is a Coordinate
        if not isinstance(other, Coordinate):
            raise TypeError("other must be a Coordinate")

        self_lat, self_lon = self.normalize()
        other_lat, other_lon = other.normalize()

        dlon = math.radians(other_lon - self_lon)
        self_lat = math.radians(self_lat)
        other_lat = math.radians(other_lat)

        x = math.sin(dlon) * math.cos(other_lat)
        y = math.cos(self_lat) * math.sin(other_lat) - math.sin(self_lat) * math.cos(
            other_lat
        ) * math.cos(dlon)
        initial_bearing = math.atan2(x, y)
        initial_bearing = math.degrees(initial_bearing)
        compass_bearing = (initial_bearing + 360) % 360
        return compass_bearing
